/* Simulation to measure BER from convolutionally encoded
   (distortion-compensated) QIM */

/* Brian Chen (bchen@alum.mit.edu)
   Massachusetts Institute of Technology
   October 1999 */

/* The purpose of publishing this software code is to educate others
   and encourage further research activity in this QIM area of
   technology.  Nothing herein, however, shall be construed as
   offering a license under any intellectual property right (such as
   patent or copyright) to use the software for any commercial
   application.  Should you be interested in using this or related
   software or technology for any commercial application, please
   contact

	Jack Turner
	Associate Director
	MIT Technology Licensing Office
	77 Massachusetts Ave., Rm. NE25-230
	Cambrige, MA  02139
	(617) 253-6966
	jht@mit.edu

   to discuss licensing opportunities. */


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "conv_def.h"
#include "sim_func.h"

#define DELTA 1.0                  /* Quantization step size. */
#define L_STDM 2.0                 /* STDM spreading length. */
#define SNRX 40.0                  /* Host signal-to-noise ratio */

#define BLK_SIZE 1000              /* Length of decoding blocks */
#define NUM_BLKS 10000             /* Number of blocks */
#define NUM_DNRS 4                 /* Number of distortion-to-noise ratios */
#define MSG_FREQ 100		   /* A status message will be written every MSG_FREQ blocks. */

/* Functions defined below. */
void gen_enc( char *input, int *state, double *symbols, int num_bits,
	      double alpha );


/* (Convolutional encoder) generator polynomials */
int G[CRATE] = {0561,0753};       /* g[0] = 101 110 001,
			             g[1] = 111 101 011 */

#if FALSE
int G[CRATE] = {0463,0535,0733,0745};  /* g[0] = 100 110 011,
					  g[1] = 101 011 101,
				          g[2] = 111 011 011,
				          g[3] = 111 100 101 */
#endif

double dnr_meas, Ds, Pn, sig_x;


void main()
{
  char u[BLK_SIZE+MEM];              /* Input to convolutional encoder */
  double x[CRATE*(BLK_SIZE+MEM)];    /* Coded symbols, channel input */
  double y[CRATE*(BLK_SIZE+MEM)];    /* Channel output */
  double qy[CRATE][NUM_VCTRS];       /* Quantized channel outputs. */
  char u_hat[BLK_SIZE];              /* Viterbi algorithm bit estimates */
  char cvect[2][NUM_STATES];         /* Code vectors corresponding to state
				        transitions */
  double PM[2][NUM_STATES];          /* Path metrics (two banks) */
  int surv[NUM_REG][2][NUM_STATES];  /* Surviving paths (two banks). Bank i is
					concatenation
					surv[NUM_REG-1][i]...surv[0][i] */
  char bank=0;                       /* Indicator for which bank is
					most recently updated. */
  double dnr, dnr_l, sig_n;
  int state=0, newState;
  double alpha;                      /* Distortion compensation parameter */

  int bit_err;                       /* Bit-error counter */
  double p_b;
  int blk;                           /* Block number */

  char b;
  int i,j,k,s;
  double d;

  FILE *fout, *fstatus;              /* Files for output and for
					reporting execution status. */

  /* DNR in dB */
  double dnr_array[NUM_DNRS]={2.0, 2.5, 3.0, 3.5};


  /* Open files. */
  fout = fopen("results.txt","a");
  /* Use fstatus=stdout to send status messages to screen.  Otherwise,
  use fopen() to write status messages to file. */
  fstatus = fopen("simulate-stat.txt","w");
#if FALSE
  fstatus = stdout;
#endif

  fprintf(fout, "Generator polynomials (octal) = ( " );
  for (i=0; i<CRATE-1; i++)
    fprintf(fout, "%o, ", G[i] );
  fprintf(fout, "%o )\n", G[CRATE-1] );
  fprintf(fout, "M=%d, Rconv=1/%d, L_STDM=%1.0f\n", MEM, CRATE, L_STDM );
  fprintf(fout, "DELTA=%2.1f, SNR_x=%2.1f dB\n", DELTA, SNRX );
  fprintf(fout, "(%1.1e bits)\n\n", (float) NUM_BLKS*BLK_SIZE);

  fprintf(fout," DNR \talpha \tbit_err \tP_b \t\tDNR_1 \tDNR_2\n");
  fprintf(fout," === \t===== \t======= \t=== \t\t===== \t=====\n");
  fflush(fout);


  /* Pre-compute code vectors corresponding to state transitions.  Only states
     (j<<1)&(NUM_STATES-1) and ((j<<1)^(NUM_STATES-1))+1 can
     transition to state j.  The input bit is the MEM-th MSB of j. */
  for (j=0; j<NUM_STATES; j++)
    {
      b = (char) ((j>>(MEM-1)) & 1);

      i = (j<<1)&(NUM_STATES-1);
      convEnc( b, i, &cvect[0][j], &newState, CRATE, MEM, G );
      i++;
      convEnc( b, i, &cvect[1][j], &newState, CRATE, MEM, G );
    }


  /* Loop through all DNRs. */
  for(s=0; s<NUM_DNRS; s++)
    {
      /* Initialize distortion and noise power registers. */
      Ds=0;
      Pn=0;

      /* Compute sig_n, given DNR and D_s=DELTA^2/(12*L_STDM). */
      dnr = dnr_array[s];
      dnr_l = exp( (dnr/10)*log(10) );    /* DNR on linear (non-dB) scale:
					      exp( x*log(10) ) = 10^(x) */

      alpha = (L_STDM*dnr_l) / ((L_STDM*dnr_l)+1);
      sig_n = sqrt( (DELTA*DELTA)/(12.0*L_STDM) / dnr_l );

      /* Compute sig_x, given sig_n. */
      sig_x = sig_n * sqrt( exp( (SNRX/10)*log(10) ) );

      /* Initialize error counter */
      bit_err=0;

      fprintf(fstatus, "DNR = %2.1f dB", dnr);
 
      for (blk=0; blk<NUM_BLKS; blk++)
	{
	  if ( MSG_FREQ*(blk/MSG_FREQ) == blk )
	  {
	    fprintf(fstatus,"\tProcessing block %d of %d...\n", blk+1, NUM_BLKS);
	    fflush(fstatus);
	  }

	  /* Generate information bits and encode data. */
	  for (i=0; i<BLK_SIZE; i++)
	    u[i] = ( (1024 & rand()) > 0 );

	  /* Tail bits */
	  for (i=BLK_SIZE; i<BLK_SIZE+MEM; i++ )
	    u[i] = 0;

	  gen_enc( &u[0], &state, &x[0], BLK_SIZE+MEM, alpha );

	  /* Pass through channel and account for puncturing */
	  gaussian_ch(x,y,sig_n,(BLK_SIZE+MEM)*CRATE);
	  for (i=0; i<BLK_SIZE+MEM; i++)
	    for (j=0; j<CRATE; j++)
	      Pn += (y[i*CRATE+j]-x[i*CRATE+j])*(y[i*CRATE+j]-x[i*CRATE+j]);

	  /* Initialize decoder to start from zero state. */
	  PM[0][0] = 0;
	  PM[1][0] = 0;
	  for (i=1; i<NUM_STATES; i++)
	    {
	      PM[0][i] = 1000.;
	      PM[1][i] = 1000.;
	    }

	  /* Initialization.  Load up decoder with first PATH_MEM
	     symbols, ignoring output bits. */
	  state = 0;
	  for (i=0; i<PATH_MEM; i++)
	    {
	      /* Quantize y using the quantizer corresponding to each
		 codevector. */
	      for (j=0; j<NUM_VCTRS; j++)
		for (k=0; k<CRATE; k++)
		  {
		    b = (j>>k) & 1;
		    b = (2*b)-1;        /* convert {0,1} to {-1,1} */
		    d = ((double) b) * DELTA/4.0;   /* d = +/- DELTA/4 */
		    qy[k][j] = dith_quant( y[CRATE*i+k], DELTA/alpha,
					   d/alpha );
		  }

	      viterbi_cont(&y[CRATE*i], qy, u_hat, PM, surv, &bank, 1, cvect);
	    }

	  /* Decode BLK_SIZE+MEM-PATH_MEM bits */
	  for (i=PATH_MEM; i<BLK_SIZE+MEM; i++)
	    {
	      /* Quantize y using the quantizer corresponding to each
		 codevector. */
	      for (j=0; j<NUM_VCTRS; j++)
		for (k=0; k<CRATE; k++)
		  {
		    b = (j>>k) & 1;
		    b = (2*b)-1;        /* convert {0,1} to {-1,1} */
		    d = ((double) b) * DELTA/4.0;   /* d = +/- DELTA/4 */
		    qy[k][j] = dith_quant( y[CRATE*i+k], DELTA/alpha,
					   d/alpha );
		  }

	      viterbi_cont(&y[CRATE*i], qy, &u_hat[i-PATH_MEM], PM,
			   surv, &bank, 1, cvect);
	    }

	  /* Last 32-MEM bits in packet should come from path
	     terminating at state 0. */
	  surv[0][bank][0] >>= MEM;       /* Shift out terminating zeros */

	  /* Last 32-MEM bits from surv[0] */
	  for (i=BLK_SIZE-1; i>BLK_SIZE+MEM-33; i--)
	    {
	      u_hat[i] = surv[0][bank][0]&1;
	      surv[0][bank][0] >>= 1;
	    }

	  /* 32 bits from each of surv[1],...,surv[NUM_REG] */
	  i = BLK_SIZE+MEM-33;
	  for (j=1; j<NUM_REG; j++)
	    for (k=0; k<32; k++)
	      {
		u_hat[i] = surv[j][bank][0]&1;
		surv[j][bank][0] >>= 1;
		i--;
	      }

	  /* Count bit-errors .*/
	  for (i=0; i<BLK_SIZE; i++)
	    {
	      if (u_hat[i] != u[i])
		bit_err++;
	    }
	}

      p_b = (double) bit_err / (double) (NUM_BLKS*BLK_SIZE);
      dnr_meas = 10*log10(Ds/L_STDM/Pn);
      fprintf(fout," %2.1f \t%1.4f \t%7d \t%1.2e \t%2.1f",
	      dnr, alpha, bit_err, p_b, dnr_meas);
      dnr_meas =
	10*log10
	( (Ds/L_STDM) / (sig_n*sig_n*(BLK_SIZE+MEM)*CRATE*NUM_BLKS) );
      fprintf(fout," \t%2.1f\n", dnr_meas);
      fflush(fout);
    }

  fclose(fstatus);
  fclose(fout);
}


void gen_enc( char *input, int *state, double *x_comp, int num_bits,
	      double alpha )
/* Generate composite signal values (x_comp) from num_bits information bits
   (input).  The current state of the convolutional encoder is state.
   The distortion compensation parameter is alpha. */
{
  int i, c, k, newState;
  char convData;
  double x_host, d;
  double z=0;

  for (i=0; i<num_bits; i++)
    {
      convEnc( input[i], *state, &convData, &newState, CRATE, MEM, G);
      *state = newState;

      for (k=0; k<CRATE; k++)
	{
	  gaussian_ch(&z,&x_host,sig_x,1);  /* 0-mean, Gauss. host */
	  c = (convData>>k) & 1;            /* kth LSB is c_k */
	  c = (2*c)-1;                      /* convert {0,1} to {-1,1} */
	  d = ((double) c) * DELTA/4.0;     /* d = +/- DELTA/4 */

	  /* Quantize and distortion compensate */
	  x_comp[CRATE*i+k] = dith_quant( alpha*x_host, DELTA, d );
	  x_comp[CRATE*i+k] += (1.0-alpha)*x_host;
	  Ds += (x_comp[CRATE*i+k]-x_host)*(x_comp[CRATE*i+k]-x_host);
	}
    }
}
