/* Viterbi decoder using inner product metric for infinite stream.  Path
   memory is 3*NUM_REG. */

/* Brian Chen (bchen@alum.mit.edu)
   Massachusetts Institute of Technology
   October 1999 */


#include "stdio.h"
#include <math.h>
#include "conv_def.h"


void viterbi_cont(double *input, double qi[][NUM_VCTRS], char *output,
		  double PM[2][NUM_STATES], int surv[NUM_REG][2][NUM_STATES],
		  char *bank, int num_bits, char cvect[2][NUM_STATES])
/* input       --> Array of received symbols
   qi          --> Array of quantized received symbols.  qi[][i] are obtained
                   using quantizer associated with codevector i.
   output      --> Array of the estimated information bits
   PM          --> Two banks to store the path metrics.
   surv        --> Two banks to store the survivor.  Each bank is the
                   concatenation surv[NUM_REG-1][bank]-...-surv[0][bank].
   bank        --> Indicator of which bank stores the most recently updated
                   metrics and survivor paths.
   num_bits    --> Number of information bits to be decoded, i.e., number
                   of steps to run Viterbi algorithm.
   cvect       --> Codevectors associated with each state transition.
                   cvect[i][j] is the code vector associated with the
		   transition from state-si to state-j.
*/
{
  char bank0=*bank;            /* indicator of which PM bank */
  char bank1=(!bank0);
  double bm[NUM_VCTRS];        /* Branch metrics */
  int s0, s1;                 /* Origin of competing paths */
  double pm0, pm1;             /* Metrics of competing paths */
  double PM_min;               /* Minimum metric value and index */

  int i, j, n, u, shift;
  double d;


  for (n=0; n<num_bits; n++)
    {
      /* Compute branch metrics for each possible code vector by taking
	 Euclidean distance between nearest reconstruction vector and
	 received vector. */
      for (i=0; i<NUM_VCTRS; i++)
	{
	  bm[i] = 0;
	  for (j=0; j<CRATE; j++)
	    {
	      d = (input[CRATE*n+j]-qi[CRATE*n+j][i]);
	      bm[i] += d*d;
	    }
	}

      /* Update path metrics and survivor paths */
      for (j=0; j<NUM_STATES; j++)
	{ 
	  /* The shift register is LEFT to RIGHT, so only states
	     (j<<1)&(NUM_STATES-1) and ((j<<1)&(NUM_STATES-1))+1 can
	     transition to state j.  The input bit u is the MEM-th MSB
	     of j. */
	  s0 = (j<<1) & (NUM_STATES-1);
	  s1 = s0 ^ 1;
	  u = (j>>(MEM-1)) & 1;

	  /* Compute incoming metrics */
	  pm0 = bm[ cvect[0][j] ] + PM[bank0][s0];
	  pm1 = bm[ cvect[1][j] ] + PM[bank0][s1];

	  /* Compare competing metrics, and keep lowest.  If path metric is
	     lowest, write outgoing bit of SURVIVOR shift register to
	     output. */
	  if (pm0<=pm1)              /* Choose path from s0 */
	    {
	      PM[bank1][j] = pm0;

	      /* Shift input bit into survivor shift register. */
	      for (i=NUM_REG-1; i>0; i--)
		{
		  shift = (surv[i-1][bank0][s0]>>31) & 1;
		  surv[i][bank1][j] = (surv[i][bank0][s0]<<1) ^ shift;
		}
	      surv[0][bank1][j] = (surv[0][bank0][s0]<<1) ^ u;

	      /* If path metric is lowest, write outgoing bit to
		 output. */
	      if ( !j || pm0<PM_min )
		{
		  output[n] = (surv[NUM_REG-1][bank0][s0]>>31) & 1;
		  PM_min = pm0;
		}
	    }
	  else                       /* Choose path from s1 */
	    {
	      PM[bank1][j] = pm1;

	      /* Shift input bit into survivor shift register. */
	      for (i=NUM_REG-1; i>0; i--)
		{
		  shift = (surv[i-1][bank0][s1]>>31) & 1;
		  surv[i][bank1][j] = (surv[i][bank0][s1]<<1) ^ shift;
		}
	      surv[0][bank1][j] = (surv[0][bank0][s1]<<1) ^ u;

	      /* If path metric is lowest, write outgoing bit to
		 output. */
	      if ( !j || pm1<PM_min )
		{
		  output[n] = (surv[NUM_REG-1][bank0][s1]>>31) & 1;
		  PM_min = pm1;
		}
	    }
	}


      /* Toggle bank0 and bank1 between 0 and 1 */
      bank0 = bank0^1;
      bank1 = bank1^1;
    }


  /* Return the indicator of which bank contains the most recently updated
     metrics and survivors.  Also, subtract PM_min from each path metric to
     avoid possible overflow. */
  *bank = bank0;
  for (j=0; j<NUM_STATES; j++ )
    PM[bank0][j] -= PM_min;
}
